/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "gwf_source_tools.h"
#include "gwf_project.h"
#include <ctype.h> /* For toupper */
#include <string.h> /* For strncmp & strchr */
#include <gtk/gtk.h>

static void		get_widname_parts		(gchar *name,gint *nparts,gchar ***parts);

static gchar*		name_to_func			(gchar *name);
static gchar*		name_to_struct			(gchar *name);
static void		name_to_def 			(gchar *name,gchar **part1,gchar **part2);
static gchar*		name_to_var			(gchar *name);
static gint		find_text			(gchar *src,gchar *text);



gchar*
gwf_source_get_widname_as_func(GwfProject *project)
{
  g_return_val_if_fail(project != NULL,NULL);
  return name_to_func(gwf_project_get_widname(project));
}

gchar*
gwf_source_get_widname_as_struct(GwfProject *project)
{
  g_return_val_if_fail(project != NULL,NULL);
  return name_to_struct(gwf_project_get_widname(project));
}

void
gwf_source_get_widname_as_def(GwfProject *project,gchar **part1,gchar **part2)
{
  g_return_if_fail(project != NULL);
  name_to_def(gwf_project_get_widname(project),part1,part2);
}

gchar*
gwf_source_get_widname_as_var(GwfProject *project)
{
  g_return_val_if_fail(project != NULL,NULL);
  return name_to_var(gwf_project_get_widname(project));
}

gchar*
gwf_source_get_parent_as_func(GwfProject *project)
{
  g_return_val_if_fail(project != NULL,NULL);
  return name_to_func(gwf_project_get_prntname(project));
}

gchar*
gwf_source_get_parent_as_struct(GwfProject *project)
{
  g_return_val_if_fail(project != NULL,NULL);
  return name_to_struct(gwf_project_get_prntname(project));
}

void
gwf_source_get_parent_as_def(GwfProject *project,gchar **part1,gchar **part2)
{
  g_return_if_fail(project != NULL);
  name_to_def(gwf_project_get_prntname(project),part1,part2);
}

gchar*
gwf_source_get_parent_as_var(GwfProject *project)
{
  g_return_val_if_fail(project != NULL,NULL);
  return name_to_var(gwf_project_get_prntname(project));
}

void
gwf_source_add_tabs_before (GString *src,gchar *text,gint maxlen)
{
  gint		len;

  g_return_if_fail(src != NULL);
  g_return_if_fail(text != NULL);

  maxlen += 8 - (maxlen % 8);
  len = strlen(text);

  if (len < maxlen)
    g_string_append(src,"\t");

  for (len += 8 - (len % 8); len < maxlen ; len += 8)
    g_string_append(src,"\t");

  g_string_append(src,text);
}

void
gwf_source_add_tabs_after (GString *src,gchar *text,gint maxlen)
{
  gint		len;

  g_return_if_fail(src != NULL);
  g_return_if_fail(text != NULL);

  maxlen += 8 - (maxlen % 8);
  g_string_append(src,text);
  len = strlen(text);

  if (len < maxlen)
    g_string_append(src,"\t");

  for (len += 8 - (len % 8); len < maxlen ; len += 8)
    g_string_append(src,"\t");
}

void
gwf_source_add_tabs_before_printf	(GString *src,gint maxlen,gchar *format,...)
{
  gchar *buffer;
  va_list args;

  g_return_if_fail(src != NULL);
  g_return_if_fail(format != NULL);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  gwf_source_add_tabs_before(src,buffer,maxlen);

  g_free(buffer);
}

void
gwf_source_add_tabs_after_printf	(GString *src,gint maxlen,gchar *format,...)
{
  gchar *buffer;
  va_list args;

  g_return_if_fail(src != NULL);
  g_return_if_fail(format != NULL);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  gwf_source_add_tabs_after(src,buffer,maxlen);

  g_free(buffer);
}

void
gwf_source_add_tabs_before_until (GString *src,gchar *text,gint until)
{
  gint		pos;
  gint		nspaces;

  g_return_if_fail(src != NULL);
  g_return_if_fail(text != NULL);

  pos = gwf_source_get_real_line_length(src->str,src->len);
  pos += (pos % 8 != 0) ? 8 - pos % 8 : 8;
  g_string_append_c(src,'\t');

  for (nspaces = until - pos ; nspaces > 8  ; nspaces -= 8)
    g_string_append_c(src,'\t');

  for (  ; nspaces > 0 ; nspaces--)
    g_string_append_c(src,' ');

  g_string_append(src,text);
}

void
gwf_source_add_tabs_after_until (GString *src,gchar *text,gint until)
{
  gint		pos;
  gint		nspaces;

  g_return_if_fail(src != NULL);
  g_return_if_fail(text != NULL);

  g_string_append(src,text);

  pos = gwf_source_get_real_line_length(src->str,src->len);

  pos += (pos % 8 != 0) ? 8 - pos % 8 : 8;
  g_string_append_c(src,'\t');

  for (nspaces = until - pos ; nspaces > 8  ; nspaces -= 8)
    g_string_append_c(src,'\t');

  for (  ; nspaces > 0 ; nspaces--)
    g_string_append_c(src,' ');
}

void
gwf_source_add_tabs_before_until_printf	(GString *src,gint until,gchar *format,...)
{
  gchar *buffer;
  va_list args;

  g_return_if_fail(src != NULL);
  g_return_if_fail(format != NULL);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  gwf_source_add_tabs_before_until(src,buffer,until);

  g_free(buffer);
}

void
gwf_source_add_tabs_after_until_printf	(GString *src,gint until,gchar *format,...)
{
  gchar *buffer;
  va_list args;

  g_return_if_fail(src != NULL);
  g_return_if_fail(format != NULL);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  gwf_source_add_tabs_after_until(src,buffer,until);

  g_free(buffer);
}


gint
gwf_source_insert_tabs_before (GString *src,gchar *text,gint pos,gint maxlen)
{
  gint		len;

  g_return_val_if_fail(src != NULL,-1);
  g_return_val_if_fail(text != NULL,-1);

  maxlen += 8 - (maxlen % 8);
  len = strlen(text);

  if (len < maxlen)
    g_string_insert(src,pos++,"\t");

  if (len % 8 != 0)
    len += 8 - (len % 8);

  for ( ; len < maxlen ; len += 8)
    g_string_insert(src,pos++,"\t");

  g_string_insert(src,pos,text);
  pos += strlen(text);

  return pos;
}

gint
gwf_source_insert_tabs_after (GString *src,gchar *text,gint pos,gint maxlen)
{
  gint		len;

  g_return_val_if_fail(src != NULL,-1);
  g_return_val_if_fail(text != NULL,-1);

  maxlen += 8 - (maxlen % 8);
  g_string_insert(src,pos,text);
  len = strlen(text);
  pos += len;

  if (len < maxlen)
    g_string_insert(src,pos++,"\t");

  if (len % 8 != 0)
    len += 8 - (len % 8);

  for ( ; len < maxlen ; len += 8)
    g_string_insert(src,pos++,"\t");

  return pos;
}

gint
gwf_source_insert_tabs_before_printf (GString *src,gint pos,gint maxlen,gchar *format,...)
{
  gchar *buffer;
  va_list args;
  gint	result;

  g_return_val_if_fail(src != NULL,-1);
  g_return_val_if_fail(format != NULL,-1);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  result = gwf_source_insert_tabs_before(src,buffer,pos,maxlen);

  g_free(buffer);

  return result;
}

gint
gwf_source_insert_tabs_after_printf (GString *src,gint pos,gint maxlen,gchar *format,...)
{
  gchar *buffer;
  va_list args;
  gint	result;

  g_return_val_if_fail(src != NULL,-1);
  g_return_val_if_fail(format != NULL,-1);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  result = gwf_source_insert_tabs_after(src,buffer,pos,maxlen);

  g_free(buffer);

  return result;
}

gint
gwf_source_insert_tabs_before_until (GString *src,gchar *text,gint pos,gint until)
{
  gint		len;
  gint		i;
  gint		nspaces;

  g_return_val_if_fail(src != NULL,0);
  g_return_val_if_fail(text != NULL,0);

  pos = gwf_source_insert(src,pos,text);
  pos = gwf_source_insert(src,pos,"\t");

  len = gwf_source_get_real_line_length(src->str,src->len);
  if (len % 8 != 0)
    len += 8 - len % 8;

  nspaces = until - len;

  for ( i = nspaces / 8 ; i > 0  ; i--, nspaces -= 8)
    pos = gwf_source_insert(src,pos,"\t");

  for (  ; nspaces > 0 ; nspaces--)
    pos = gwf_source_insert(src,pos," ");

  return pos;
}

gint
gwf_source_insert_tabs_after_until (GString *src,gchar *text,gint pos,gint until)
{
  gint		len;
  gint		i;
  gint		nspaces;

  g_return_val_if_fail(src != NULL,0);
  g_return_val_if_fail(text != NULL,0);

  pos = gwf_source_insert(src,pos,"\t");

  len = gwf_source_get_real_line_length(src->str,src->len);
  if (len % 8 != 0)
    len += 8 - len % 8;

  nspaces = until - len;

  for ( i = nspaces / 8 ; i > 0  ; i--, nspaces -= 8)
    pos = gwf_source_insert(src,pos,"\t");

  for (  ; nspaces > 0 ; nspaces--)
    pos = gwf_source_insert(src,pos," ");

  pos = gwf_source_insert(src,pos,text);

  return pos;
}

gint
gwf_source_insert_tabs_before_until_printf (GString *src,gint pos,gint until,gchar *format,...)
{
  gchar *buffer;
  va_list args;
  gint	result;

  g_return_val_if_fail(src != NULL,-1);
  g_return_val_if_fail(format != NULL,-1);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  result = gwf_source_insert_tabs_before_until(src,buffer,pos,until);

  g_free(buffer);

  return result;
}

gint
gwf_source_insert_tabs_after_until_printf (GString *src,gint pos,gint until,gchar *format,...)
{
  gchar *buffer;
  va_list args;
  gint	result;

  g_return_val_if_fail(src != NULL,-1);
  g_return_val_if_fail(format != NULL,-1);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  result = gwf_source_insert_tabs_after_until(src,buffer,pos,until);

  g_free(buffer);

  return result;
}

gint
gwf_source_insert(GString *src,gint pos,gchar *text)
{
  g_return_val_if_fail(src != NULL,-1);
  g_return_val_if_fail(text != NULL,-1);

  g_string_insert(src,pos,text);
  return (strlen(text) + pos);
}

gint
gwf_source_insert_printf(GString *src,gint pos,gchar *format,...)
{
  gchar *buffer;
  va_list args;
  gint	result;

  g_return_val_if_fail(src != NULL,-1);
  g_return_val_if_fail(format != NULL,-1);

  va_start (args, format);
  buffer = g_strdup_vprintf (format, args);
  va_end (args);

  result = gwf_source_insert(src,pos,buffer);

  g_free(buffer);

  return result;
}

gint
gwf_source_get_line_length (gchar *line,gint pos)
{
  gint		len;
  gchar		*tmp;

  tmp = line + pos;

  for (len = 0, tmp = line + pos ; (tmp >= line) && (*tmp != '\n') ; tmp--,len++) ;
  for (tmp = line + pos ; (*tmp != '\n') && (*tmp != '\0') ; tmp++,len++) ;

  return (len == 0) ? 0 : len - 1;
}

gint
gwf_source_get_real_line_length (gchar *line,gint pos)
{
  gint		len;
  gchar		*tmp;

  for (tmp = line + pos ; (tmp > line) && (*tmp != '\n') ; tmp--) ;

  for (len = 0, tmp++ ; (*tmp != '\n') && (*tmp != '\0') ; tmp++)
    len += (*tmp == '\t') ? 8 - len % 8 : 1;

  return (len == 0) ? 0 : len;
}

gint
gwf_source_find_text (GString *src,gchar *text)
{
  return find_text(src->str,text);
}

gint
gwf_source_find_text_from_pos(GString *src,gint pos,gchar *text)
{
  return (find_text(src->str + pos, text) + pos);
}


static void
get_widname_parts(gchar *name,gint *nparts,gchar ***parts)
{
  gint		len,i,j;

  len = strlen(name);

  *nparts = 1;

  for ( i = 1 ; i < len ; i++)
    if (isupper(name[i])) (*nparts)++;

  *parts = g_malloc(sizeof(gchar*) * *nparts);
  for ( i = 0 ; i < *nparts ; i++)
  {
    for ( j=1 ; !(isupper(name[j])) && (name[j] != '\0'); j++) ;

    (*parts)[i] = g_strndup(name, j);
    toupper((*parts)[i][0]);
    name += j;
  }
}

static gchar*
name_to_func(gchar *name)
{
  int		nparts,i;
  gchar 	**parts;
  gchar		*tmp,*tmp2;

  g_return_val_if_fail(name != NULL,NULL);

  get_widname_parts(name,&nparts,&parts);

  tmp = parts[0];
  for (i = 1 ; i < nparts ; i++)
  {
    tmp2 = tmp;
    tmp = g_strdup_printf("%s_%s",tmp,parts[i]);
    g_free(tmp2);
    g_free(parts[i]);
  }
  g_free(parts);

  g_strdown(tmp);

  return tmp;
}

static gchar*
name_to_struct(gchar *name)
{
  g_return_val_if_fail(name != NULL,NULL);

  return g_strdup(name);
}

static void
name_to_def (gchar *name,gchar **part1,gchar **part2)
{
  int		nparts,i;
  gchar 	**parts;
  gchar		*tmp,*tmp2;

  g_return_if_fail(name != NULL);

  get_widname_parts(name,&nparts,&parts);

  g_strup(parts[0]);
  *part1 = parts[0];
  if (nparts == 1)
  {
    *part2 = NULL;
    return;
  }

  tmp = parts[1];
  for (i = 2 ; i < nparts ; i++)
  {
    tmp2 = tmp;
    tmp = g_strdup_printf("%s_%s",tmp,parts[i]);
    g_free(tmp2);
    g_free(parts[i]);
  }
  g_free(parts);

  g_strup(tmp);
  *part2 = tmp;
}

static gchar*
name_to_var (gchar *name)
{
  int		nparts,i;
  gchar 	**parts;
  gchar		*tmp,*tmp2;

  g_return_val_if_fail(name != NULL,NULL);

  get_widname_parts(name,&nparts,&parts);

  g_strdown(parts[0]);

  tmp = (nparts == 1) ? parts[0] : parts[1];

  for (i = 2 ; i < nparts ; i++)
  {
    tmp2 = tmp;
    tmp = g_strdup_printf("%s_%s",tmp,parts[i]);
    g_free(tmp2);
    g_free(parts[i]);
  }
  g_free(parts);

  g_strdown(tmp);

  return tmp;
}

static gint
find_text(gchar *src,gchar *text)
{
  gchar		*tmp;

  tmp = strchr(src,*text);
  for ( ; tmp && (strncmp(tmp,text,strlen(text)) != 0) ;)
   tmp = strchr(tmp+1,*text);

  return (tmp) ? (gint)(tmp - src) : -1;
}

